from typing import Optional, List, Dict
from ontobio.model.similarity import AnnotationSufficiency
from ontobio.vocabulary.upper import HpoUpperLevel
from ontobio.sim.api.interfaces import InformationContentStore
import numpy as np
from statistics import mean


class AnnotationScorer:
    """
    Computes the annotation sufficiency scores as described
    by https://zenodo.org/record/834091#.W8ZnCxhlCV4
    """

    def __init__(self, ic_store: InformationContentStore):
        self.ic_store = ic_store

    def get_annotation_sufficiency(
            self,
            profile: List[str],
            negated_classes: List[str],
            categories: Optional[List] = None,
            negation_weight: Optional[float] = .25,
            category_weight: Optional[float] = .5) -> AnnotationSufficiency:
        """
        Given a list of individuals, return the simple, scaled, and categorical scores
        """

        if categories is None:
            categories = [enum.value for enum in HpoUpperLevel]
        ic_map = self.ic_store.get_profile_ic(profile + negated_classes)

        # Simple score is the weighted average of the present and
        # explicitly stated negative/absent phenotypes
        #
        # Note that we're deviating from the publication
        # to match the reference java implementation where
        # mean_max_ic is replaced with max_max_ic:
        # https://github.com/owlcollab/owltools/blob/452b4a/
        # OWLTools-Sim/src/main/java/owltools/sim2/AbstractOwlSim.java#L1038
        simple_score = self._get_simple_score(
            profile, negated_classes, self.ic_store.statistics.mean_mean_ic,
            self.ic_store.statistics.max_max_ic, self.ic_store.statistics.mean_sum_ic,
            negation_weight, ic_map
        )

        categorical_score = self._get_categorical_score(
            profile, negated_classes, categories,
            negation_weight, ic_map
        )
        scaled_score = self._get_scaled_score(
            simple_score, categorical_score, category_weight)

        return AnnotationSufficiency(
            simple_score=simple_score,
            scaled_score=scaled_score,
            categorical_score=categorical_score
        )

    def _get_simple_score(self,
                          profile: List[str],
                          negated_classes: List[str],
                          bg_mean_pic: float,
                          bg_mean_max_pic: float,
                          bg_mean_sum_pic: float,
                          negation_weight: Optional[float] = .25,
                          ic_map: Optional[Dict[str, float]] = None) -> float:
        """
        Simple score is the average of the relative
        mean ic, max ic, and sum ic (relative to global stats)

        :param ic_map: dictionary of class - information content mappings
        :param bg_mean_pic: the average of the average IC in
                            the background profile annotations
        :param bg_mean_max_pic: max IC annotated to the background set of profiles
        :param bg_mean_sum_pic: Average of the profile sum IC in background set
        :param negation_weight: Average of the profile sum IC in background set
        :param ic_map: Average of the profile sum IC in background set
        :return: simple score (float)
        """
        if ic_map is None:
            ic_map = self.ic_store.get_profile_ic(profile + negated_classes)

        pos_map = {cls: ic for cls, ic in ic_map.items() if cls in profile}
        neg_map = {cls: ic for cls, ic in ic_map.items() if cls in negated_classes}

        mean_ic = mean(pos_map.values()) if len(profile) > 0 else 0
        max_ic = max(pos_map.values()) if len(profile) > 0 else 0
        sum_ic = sum(pos_map.values()) if len(profile) > 0 else 0

        if len(negated_classes) > 0:
            weighted_ic = [ic * negation_weight for ic in neg_map.values()]
            mean_ic = max([np.average([mean_ic, mean(neg_map.values())],
                                      weights=[1, negation_weight]),
                           mean_ic])
            max_ic = max([max_ic] + weighted_ic)
            sum_ic = sum_ic + sum(weighted_ic)

        return mean([
            min([mean_ic / bg_mean_pic, 1.0]),
            min([max_ic / bg_mean_max_pic, 1.0]),
            min([sum_ic / bg_mean_sum_pic, 1.0])
        ])

    @staticmethod
    def _get_scaled_score(
            simple_score: float,
            categorical_score: float,
            category_weight: Optional[float] = .5) -> float:
        """
        Scaled score is the weighted average of the simple score and
        categorical score
        """
        return np.average(
            [simple_score, categorical_score], weights=[1, category_weight]
        )

    def _get_categorical_score(
            self,
            profile: List,
            negated_classes: List,
            categories: List,
            negation_weight: Optional[float] = 1,
            ic_map: Optional[Dict[str, float]] = None) -> float:
        """
        The average of the simple scores across a list of categories
        """

        if ic_map is None:
            ic_map = self.ic_store.get_profile_ic(profile + negated_classes)

        scores = []

        for cat in categories:
            if cat not in self.ic_store.category_statistics:
                raise ValueError("statistics for {} not indexed".format(cat))

            pos_profile = [cls for cls in profile
                           if cls in self.ic_store.category_statistics[cat].descendants]
            neg_profile = [cls for cls in negated_classes
                           if cls in self.ic_store.category_statistics[cat].descendants]

            # Note that we're deviating from the publication
            # to match the reference java implementation where
            # mean_max_ic is replaced by max_max_ic
            scores.append(self._get_simple_score(
                pos_profile, neg_profile,
                self.ic_store.category_statistics[cat].mean_mean_ic,
                self.ic_store.category_statistics[cat].max_max_ic,
                self.ic_store.category_statistics[cat].mean_sum_ic,
                negation_weight, ic_map
            ))
        return mean(scores)



from ontobio.rdfgen import relations
from ontobio.model import association

def test_relations_curie_expand():
    uri = relations.curie_to_obo_uri(association.Curie.from_str("GO:1234567"))
    assert uri == "http://purl.obolibrary.org/obo/GO_1234567"


def test_relations_curie_contract():
    curie = relations.obo_uri_to_curie("http://purl.obolibrary.org/obo/GO_1234567")
    assert curie == association.Curie(namespace="GO", identity="1234567")




from ontobio.rdfgen.gocamgen.gocamgen import AssocGoCamModel
from ontobio.rdfgen.gocamgen.filter_rule import AssocFilter, FilterRule, get_filter_rule
from ontobio.rdfgen.gocamgen.collapsed_assoc import extract_properties
from ontobio.rdfgen.gocamgen.errors import GocamgenException, GeneErrorSet
from ontobio.io import gpadparser
from ontobio.io.assocparser import AssocParserConfig
from ontobio.io.entityparser import GpiParser
from ontobio.model.association import GoAssociation
from ontobio.ontol_factory import OntologyFactory
from ontobio.util.go_utils import GoAspector
import argparse
import logging
import requests
from requests.exceptions import ConnectionError
import gzip
import time
import click
from os import path
from typing import List
# from abc import ABC, abstractmethod
from rdflib.graph import ConjunctiveGraph
from rdflib.store import Store
from rdflib import plugin

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")

parser = argparse.ArgumentParser()
parser.add_argument('-g', '--gpad_file', help="Filepath of GPAD source with annotations to model", required=True)
parser.add_argument('-s', '--specific_gene', help="If specified, will only translate model for annotations "
                                                  "to this specific gene")
parser.add_argument('-o', '--ontology', help="Path to GO-LEGO ontology file")
parser.add_argument('-n', '--max_model_limit', help="Only translate specified number of models. Mainly for testing.")
parser.add_argument('-m', '--mod', help="MOD rules to follow for filtering and translating.")
parser.add_argument('-d', '--output_directory', help="Directory to output model ttl files to")
parser.add_argument('-r', '--report', help="Generate report", action="store_const", const=True)
parser.add_argument('-N', '--nquads', help="Filepath to write model file in N-Quads format")

# GoCamInputHandler


class GoCamBuilder:
    def __init__(self, parser_config: AssocParserConfig, modelstate=None):
        self.config = parser_config
        self.aspector = GoAspector(self.config.ontology)
        self.store = plugin.get('IOMemory', Store)()
        self.errors = GeneErrorSet()  # Errors by gene ID
        self.gpi_entities = self.parse_gpi(parser_config.gpi_authority_path)
        self.modelstate = modelstate

    def translate_to_model(self, gene, assocs: List[GoAssociation]):
        if gene not in self.gpi_entities:
            error_msg = "Gene ID '{}' missing from provided GPI. Skipping model translation.".format(gene)
            raise GocamgenException(error_msg)
        model_id = gene.replace(":", "_")
        model_title = self.model_title(gene)
        model = AssocGoCamModel(model_title,
                                assocs,
                                config=self.config,
                                store=self.store,
                                gpi_entities=self.gpi_entities,
                                model_id=model_id,
                                modelstate=self.modelstate)
        model.go_aspector = self.aspector  # TODO: Grab aspect from ontology node
        model.translate()

        return model

    def make_model(self, gene, annotations: List[GoAssociation], output_directory=None, nquads=False):
        # All these retry shenanigans are to prevent mid-run crashes due to an external resource simply blipping
        # out for a second.
        retry_count = 0
        retry_limit = 5
        while True:
            try:
                start_time = time.time()
                model = self.translate_to_model(gene, annotations)
                # add_to_conjunctive_graph(model, conjunctive_graph)
                if nquads:
                    logger.info(
                        "Model for {} added to graphstore in {} sec".format(gene, (time.time() - start_time)))
                else:
                    out_filename = "{}.ttl".format(gene.replace(":", "_"))
                    if output_directory is not None:
                        out_filename = path.join(output_directory, out_filename)
                    model.write(out_filename)
                    logger.info("Model for {} written to {} in {} sec".format(gene, out_filename,
                                                                              (time.time() - start_time)))
                for err in model.errors:
                    self.errors.add_error(gene, err)
                return model
            except GocamgenException as ex:
                self.errors.add_error(gene, ex)
            except (TimeoutError, ConnectionError) as ex:
                # This has been happening randomly and breaking full runs
                self.errors.add_error(gene, ex)
                if retry_count < retry_limit:
                    retry_count += 1
                    continue  # retry
                self.errors.add_error(gene,
                                 GocamgenException(f"Bailing on model for {gene} after {retry_count} retries"))
            break  # Done with this model. Move on to the next one.

    def make_model_and_add_to_store(self, gene, annotations):
        return self.make_model(gene, annotations, nquads=True)

    def make_model_and_write_out(self, gene, annotations, output_directory=None):
        return self.make_model(gene, annotations, output_directory=output_directory, nquads=False)

    def write_out_store_to_nquads(self, filepath):
        cg = ConjunctiveGraph(self.store)
        cg.serialize(destination=filepath, format="nquads")
        logger.info(f"Full model graphstore written out in N-Quads format to {filepath}")

    def write_report(self, report_filepath):
        with open(report_filepath, "w+") as reportf:
            for gene, errs in self.errors.errors.items():
                for ex in errs:
                    reportf.write(f"{type(ex).__name__} - {gene}: {ex}\n")

    def model_title(self, gene_id: str) -> str:
        entity = self.gpi_entities.get(gene_id)
        if entity:
            model_title = "{} ({})".format(entity["label"], gene_id)
        else:
            model_title = gene_id
        return model_title

    @staticmethod
    def extract_relations_ontology(ontology_graph):
        # TODO: This is probably all wrong - trying to reconstruct RO from broader ontology_graph
        # According to Ontobee RO top-level terms are RO:0002410 (causally related to)
        #  and RO:0002222 (temporally related to) - try reconstructing RO with these
        ro_terms = []
        for t in [
            "RO:0002410",  # causally related to
            "RO:0002222",  # temporally related to
            "RO:0002323",  # mereotopologically related to - for BFO:0000050
        ]:
            ro_terms = ro_terms + ontology_graph.descendants(t, reflexive=True)
        return ontology_graph.subontology(nodes=ro_terms)

    @staticmethod
    def parse_gpi(gpi_file):
        # {
        #    "id":"MGI:MGI:87853",
        #    "label":"a",
        #    "full_name":"nonagouti",
        #    "synonyms":[
        #       "agouti",
        #       "As",
        #       "agouti signal protein",
        #       "ASP"
        #    ],
        #    "type":"gene",
        #    "parents":[
        #
        #    ],
        #    "xrefs":[
        #       "UniProtKB:Q03288"
        #    ],
        #    "taxon":{
        #       "id":"NCBITaxon:10090"
        #    }
        # }
        if gpi_file is None:
            return None
        parser = GpiParser()
        gpi_entities = {}
        entities = parser.parse(gpi_file)
        for entity in entities:
            gpi_entities[entity['id']] = entity
        return gpi_entities


class AssocExtractor:
    def __init__(self, gpad_file, parser_config: AssocParserConfig):
        self.assocs = []
        self.gpad_parser = gpadparser.GpadParser(config=parser_config)
        with open(gpad_file) as sg:
            lines = sum(1 for line in sg)

        with open(gpad_file) as gf:
            click.echo("Making products...")
            with click.progressbar(iterable=self.gpad_parser.association_generator(file=gf, skipheader=True),
                                   length=lines) as associations:
                self.assocs = list(associations)

        self.entity_parents = self.parse_gpi_parents(parser_config.gpi_authority_path)

    def group_assocs(self):
        assocs_by_gene = {}
        for a in self.assocs:
            subject_id = str(a.subject.id)
            # If entity has parent, assign to parent entity model
            if subject_id in self.entity_parents:
                subject_id = self.entity_parents[subject_id]
            if subject_id in assocs_by_gene:
                assocs_by_gene[subject_id].append(a)
            else:
                assocs_by_gene[subject_id] = [a]
        return assocs_by_gene

    @staticmethod
    def extract_properties_from_assocs(assocs):
        new_assoc_list = []
        for a in assocs:
            new_assoc_list.append(extract_properties(a))
        return new_assoc_list

    @staticmethod
    def parse_gpi_parents(gpi_file):
        if gpi_file is None:
            return None
        parser = GpiParser()
        entity_parents = {}
        entities = parser.parse(gpi_file)
        for entity in entities:
            entity_id = entity['id']
            if len(entity['encoded_by']) > 0:
                entity_parents[entity_id] = entity['encoded_by'][0]  # There may only be one
        return entity_parents


def unzip(filepath):
    input_file = gzip.GzipFile(filepath, "rb")
    s = input_file.read()
    input_file.close()

    target = path.splitext(filepath)[0]
    logger.info("Gunzipping file: {}".format(filepath))
    with open(target, "wb") as output:
        output.write(s)
    return target


def handle_gpad_file(args_gpad_file):
    if args_gpad_file.startswith("http://"):
        gpad_file_target = args_gpad_file.split("/")[-1]
        logger.info("Downloading GPAD from {} and saving to {}".format(args_gpad_file, gpad_file_target))
        response = requests.get(args_gpad_file)
        with open(gpad_file_target, "wb") as gft:
            gft.write(response.content)
        if gpad_file_target.endswith(".gz"):
            gpad_file = unzip(gpad_file_target)
        else:
            gpad_file = gpad_file_target
    else:
        gpad_file = args.gpad_file
    return gpad_file


def parse_header(gpad_file):
    header_data = {
        "date": ""
    }
    with open(gpad_file) as gf:
        for l in gf.readlines():
            if l.startswith("!"):
                date_key = "!date: "
                if l.startswith(date_key):
                    header_data["date"] = l.split(date_key)[1].split("$")[0].strip()
            else:
                break
    return header_data


if __name__ == "__main__":
    args = parser.parse_args()

    filter_rule = get_filter_rule(args.mod)

    gpad_file = handle_gpad_file(args.gpad_file)
    relevant_header_data = parse_header(gpad_file)
    gpad_file_metadata = {
        "source_path": args.gpad_file,
        # TODO: Figure out how to get real creation date from file
        "download_date": time.ctime(path.getmtime(gpad_file)),
        "header_date": relevant_header_data["date"]
    }

    extractor = AssocExtractor(gpad_file)
    assocs_by_gene = extractor.group_assocs()
    logger.debug("{} distinct genes".format(len(assocs_by_gene)))

    ontology_graph = OntologyFactory().create(args.ontology)
    builder = GoCamBuilder(ontology_graph)

    model_count = 0
    if args.specific_gene:
        for specific_gene in args.specific_gene.split(","):
            if specific_gene not in assocs_by_gene:
                logger.error("ERROR: specific gene {} not found in filtered annotation list".format(specific_gene))
            else:
                logger.debug("{} filtered annotations to translate for {}".format(len(assocs_by_gene[specific_gene]), specific_gene))
                builder.make_model(specific_gene, annotations=assocs_by_gene[specific_gene], output_directory=args.output_directory, nquads=args.nquads)
                model_count += 1
    else:
        for gene in assocs_by_gene:
            builder.make_model(gene, annotations=assocs_by_gene[gene], output_directory=args.output_directory, nquads=args.nquads)
            model_count += 1
            if args.max_model_limit and model_count == int(args.max_model_limit):
                break

    if args.nquads:
        builder.write_out_store_to_nquads(filepath=args.nquads)

    if args.report:
        report_file_path = "{}.report".format(gpad_file)
        with open(report_file_path, "w+") as reportf:
            for k in gpad_file_metadata:
                reportf.write("{}: {}\n".format(k, gpad_file_metadata[k]))
            # TODO FilterRule().__str__() to display filters
            reportf.write("# of models generated: {}\n".format(model_count))
            for gene, errs in builder.errors.errors.items():
                for ex in errs:
                    reportf.write(f"{type(ex).__name__} - {gene}: {ex}\n")
        logger.info("Report file generated at {}".format(report_file_path))



"""
Parsers for GAF and various Association TSVs.

All parser objects instantiate a subclass of the abstract `AssocParser` object

"""

# TODO: Refactor - move some stuff out into generic parser object

# from _typeshed import NoneType
import re
import requests
import tempfile
from contextlib import closing
import subprocess
import logging
import io
import gzip
import datetime
import dateutil.parser

from dataclasses import dataclass

from collections import namedtuple, defaultdict
from typing import Callable, ClassVar, Collection, Iterable, Optional, List, Dict, Set, TypeVar, Union, Any

from ontobio import ontol
from ontobio import ecomap
from ontobio.io import parsereport
from ontobio.util.user_agent import get_user_agent
from ontobio.model import association
from ontobio.rdfgen import relations

from prefixcommons import curie_util

TAXON = 'TAXON'
ENTITY = 'ENTITY'
ANNOTATION = 'ANNOTATION'
EXTENSION = 'EXTENSION'

logger = logging.getLogger(__name__)


def write_to_file(optional_file, text):
    if optional_file:
        optional_file.write(text)


SplitLine = namedtuple("SplitLine", ["line", "values", "taxon"])
SplitLine.__str__ = lambda self: self.line
"""
This is a collection that views a gaf line in two ways: as the full line, and
as the separated by tab list of values. We also tack on the taxon.
"""

@dataclass
class ValidateResult:
    valid: bool
    original: str
    parsed: Optional[Any]
    message: str

@dataclass
class ColumnValidator:

    def cardinality(self, field: str, entity: str, minimum: int, maximum: Optional[int]) -> ValidateResult:
        items = [i for i in entity.split("|") if i != ""]
        if len(items) == 0 and 0 < minimum:
            return ValidateResult(False, "EMPTY", None, "{} must have at least {} items".format(field, minimum))
        if len(items) < minimum:
            return ValidateResult(False, entity, None, "{} must have at least {} items".format(field, minimum))
        elif maximum is not None and len(items) > maximum:
            return ValidateResult(False, entity, None, "{} cannot have more than {} items".format(field, maximum))

        return ValidateResult(True, entity, items, "")

    def validate(self, entity) -> ValidateResult:
        return ValidateResult(True, entity, entity, "")

@dataclass
class Qualifier2_1(ColumnValidator):
    valid_qualifiers = {
        "contributes_to",
        "colocalizes_with",
    }

    def validate(self, entity) -> ValidateResult:
        result = self.cardinality("Qualifier", entity, 0, 2)
        if not result.valid:
            return result

        invalid_non_not = [q for q in result.parsed if q not in self.valid_qualifiers and q != "NOT"]
        if invalid_non_not:
            result.valid = False
            result.parsed = None
            result.message = "Invalid Qualifiers: {}".format(", ".join(invalid_non_not))

        return result

@dataclass
class Qualifier2_2(ColumnValidator):

    valid_qualifiers = {
        "enables",
        "contributes_to",
        "involved_in",
        "acts_upstream_of",
        "acts_upstream_of_positive_effect",
        "acts_upstream_of_negative_effect",
        "acts_upstream_of_or_within",
        "acts_upstream_of_or_within_positive_effect",
        "acts_upstream_of_or_within_negative_effect",
        "located_in",
        "part_of",
        "is_active_in",
        "colocalizes_with"
    }

    def validate(self, entity) -> ValidateResult:
        result = self.cardinality("Qualifier", entity, 1, 2)
        if not result.valid:
            return result

        invalid_non_not = [q for q in result.parsed if q not in self.valid_qualifiers and q != "NOT"]
        if invalid_non_not:
            result.valid = False
            result.parsed = None
            result.message = "Invalid Qualifiers: {}".format(", ".join(invalid_non_not))
            return result

        if "NOT" in result.parsed and len(result.parsed) == 1:
            # If not is the only elemnent, then it's wrong
            result.valid = False
            result.parsed = None
            result.message = "`NOT` is not a Qualifier and so cannot exist by itself"
            return result

        if len(result.parsed) == 2 and "NOT" not in result.parsed:
            result.valid = False
            result.parsed = None
            result.message = "There can be only one relation entry for the Qualifier field"
            return result

        return result


@dataclass
class CurieValidator(ColumnValidator):
    def validate(self, entity):
        curie = association.Curie.from_str(entity)
        if isinstance(curie, association.Error):
            return ValidateResult(False, entity, None, curie.info)

        return ValidateResult(True, entity, curie, "")


@dataclass
class TaxonValidator(CurieValidator):

    def validate(self, entity):
        cardinality_result = self.cardinality("Taxon", entity, 1, 2)
        if not cardinality_result.valid:
            return cardinality_result

        parsed_taxons = []
        for taxa in cardinality_result.parsed:
            result = super().validate(taxa)
            if not result.valid:
                return result

            result.parsed.namespace = "NCBITaxon"
            parsed_taxons.append(result.parsed)

        return ValidateResult(True, entity, parsed_taxons, "")

@dataclass
class RuleSet:
    """
    This represents the set of GO Rules we want ontobio to run during parsing of annotations.

    If `rules` is None, all rules will be run. If `rules` is an empty set, no rules will be run.
    Otherwise, the rules set member variable is a set of integers corresponding to the rule IDs
    that will be run during parse.

    For example, if self.rules == set([6, 13, 20]) then only GORULE:0000006, GORULE:0000013, and 
    GORULE:0000020 will be run.

    Note that GORULE:0000001 will always be run, as that rule represents succesful parsing a line
    into ontobio at all.
    """
    rules: Optional[Set[int]]

    ALL: ClassVar[str] = "all"

    def __init__(self, rules: Optional[Iterable]):
        if rules is None:
            self.rules = None
        else:
            self.rules = set(rules)
    
    def should_run_rule(self, rule: int) -> bool:
        if self.rules is None:
            return True
        
        return rule in self.rules


class AssocParserConfig():
    """
    Configuration for an association parser

    rule_metadata: Dictionary of rule IDs to metadata pulled out by yamldown

    rule_sets: an Iterable of integers representing 
    """
    def __init__(self,
                 remove_double_prefixes=False,
                 ontology=None,
                 repair_obsoletes=True,
                 entity_map=None,
                 valid_taxa=None,
                 class_idspaces=None,
                 entity_idspaces=None,
                 group_idspace=None,
                 ecomap=ecomap.EcoMap(),
                 exclude_relations=None,
                 include_relations=None,
                 filter_out_evidence=None,
                 filtered_evidence_file=None,
                 gpi_authority_path=None,
                 paint=False,
                 rule_metadata=dict(),
                 goref_metadata=None,
                 group_metadata=None,
                 dbxrefs=None,
                 suppress_rule_reporting_tags=[],
                 annotation_inferences=None,
                 extensions_constraints=None,
                 rule_contexts=[],
                 rule_set=None):

        self.remove_double_prefixes=remove_double_prefixes
        self.ontology=ontology
        self.repair_obsoletes=repair_obsoletes
        self.entity_map=entity_map
        self.valid_taxa=valid_taxa
        self.class_idspaces=class_idspaces
        self.ecomap=ecomap
        self.include_relations=include_relations
        self.exclude_relations=exclude_relations
        self.filter_out_evidence = filter_out_evidence
        self.filtered_evidence_file = filtered_evidence_file
        self.gpi_authority_path = gpi_authority_path
        self.paint = paint
        self.rule_metadata = rule_metadata
        self.goref_metadata = goref_metadata
        self.group_metadata = group_metadata
        self.suppress_rule_reporting_tags = suppress_rule_reporting_tags
        self.annotation_inferences = annotation_inferences
        self.entity_idspaces = entity_idspaces
        self.extensions_constraints = AssocParserConfig._compute_constraint_subclasses(extensions_constraints, ontology)
        self.group_idspace = None if group_idspace is None else set(group_idspace)
        self.rule_contexts = rule_contexts
        # We'll say that the default None should run no rules, so let's set the rule_set to []
        # print("Rule Set is {}".format(rule_set))
        if rule_set == None:
            self.rule_set = RuleSet([])
        elif rule_set == RuleSet.ALL:
            # None here means all rules
            self.rule_set = RuleSet(None)
        else:
            self.rule_set = RuleSet(rule_set)


        # This is a dictionary from ruleid: `gorule-0000001` to title strings
        if self.exclude_relations is None:
            self.exclude_relations = []
        if self.include_relations is None:
            self.include_relations = []
        if self.filter_out_evidence is None:
            self.filter_out_evidence = []


    def _compute_constraint_subclasses(extensions_constraints, ontology):
        if extensions_constraints is None:
            return None
        # Precompute subclass closures in the extensions_constraints
        cache = dict()  # term -> [children]
        for constraint in extensions_constraints:
            terms = set()
            for term in constraint["primary_root_terms"]:
                if not term in cache and ontology is not None:
                    cache[term] = ontology.descendants(term, relations=["subClassOf"], reflexive=True)
                elif ontology is None:
                    cache[term] = [term]

                terms.update(cache[term])

            constraint["primary_terms"] = sorted(list(terms))

        return extensions_constraints

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __str__(self):
        s = "AssocParserConfig("
        attribute_values = ["{att}={val}".format(att=att, val=dict([(k, v) for k, v in value.items()][:8]) if isinstance(value, dict) else value) for att, value in vars(self).items()]
        return "AssocParserConfig({})".format(",".join(attribute_values))

class Report(object):
    """
    A report object that is generated as a result of a parse
    """

    # Levels
    """
    4  message levels
    """
    FATAL = 'FATAL'
    ERROR = 'ERROR'
    WARNING = 'WARNING'
    INFO = "INFO"

    # Warnings: TODO link to gorules
    INVALID_ID = "Invalid identifier"
    INVALID_ID_DBXREF = "ID or symbol not present in DB xrefs file"
    UNKNOWN_ID = "Unknown identifier"
    INVALID_IDSPACE = "Invalid identifier prefix"
    INVALID_QUALIFIER = "Invalid Qualifier"
    INVALID_TAXON = "Invalid taxon"
    INVALID_SYMBOL = "Invalid symbol"
    INVALID_DATE = "Invalid date"
    INVALID_ASPECT = "Invalid aspect code. Should be C, P, or F"
    UNMAPPED_ID = "Unmapped identifier"
    UNKNOWN_EVIDENCE_CLASS = "Unknown evidence class"
    OBSOLETE_CLASS = "Obsolete class"
    OBSOLETE_CLASS_NO_REPLACEMENT = "Obsolete class with no replacement"
    WRONG_NUMBER_OF_COLUMNS = "Wrong number of columns in this line"
    EXTENSION_SYNTAX_ERROR = "Syntax error in annotation extension field"
    VIOLATES_GO_RULE = "Violates GO Rule"
    RULE_PASS = "Passes GO Rule"
    INVALID_REFERENCES = "Only one reference per ID space allowed"


    def __init__(self, group="unknown", dataset="unknown", config=None):
        self.messages = []
        self.n_lines = 0
        self.n_assocs = 0
        self.skipped = 0
        self.reporter = parsereport.Report(group, dataset)
        if config is None:
            config = AssocParserConfig()
        self.config = config
        self.header = []

    def error(self, line, type, obj, msg="", taxon: str = "", rule=None):
        self.message(self.ERROR, line, type, obj, msg, taxon=taxon, rule=rule)

    def warning(self, line, type, obj, msg="", taxon: str = "", rule=None):
        self.message(self.WARNING, line, type, obj, msg, taxon=taxon, rule=rule)

    def message(self, level, line, type, obj, msg="", taxon: str = "", rule=None, dont_record=["INFO"]):
        message = {
            'level': level,
            'line': line,
            'type': type,
            'message': msg,
            'obj': obj,
            'taxon': taxon,
            'rule': rule
        }
        if not level in dont_record:
            # Only record a message if we want that
            self.messages.append(message)

        self.reporter.message(message, rule)

    def add_associations(self, associations):
        for a in associations:
            self.add_association(a)

    def add_association(self, association):
        if isinstance(association, dict) and "header" in association:
            # Then we are a header
            self.header.append(association["line"])
        else:
            self.n_assocs += 1
        # self.subjects.add(association['subject']['id'])
        # self.objects.add(association['object']['id'])
        # self.references.update(association['evidence']['has_supporting_reference'])
        # if 'taxon' in association['subject']:
        #     self.taxa.add(association['subject']['taxon']['id'])

    def report_parsed_result(self, result, output_file, evidence_filtered_file, evidence_to_filter):

        self.n_lines += 1
        if result.skipped:
            logger.debug("SKIPPING (assocparser): {}".format(result.parsed_line))
            self.skipped += 1
        else:
            self.add_associations(result.associations)
            if result.evidence_used not in evidence_to_filter:
                write_to_file(evidence_filtered_file, result.parsed_line + "\n")

    def short_summary(self):
        return "Parsed {} assocs from {} lines. Skipped: {}".format(self.n_assocs, self.n_lines, self.skipped)

    def to_report_json(self):
        """
        Generate a summary in json format
        """

        return self.reporter.json(self.n_lines, self.n_assocs, self.skipped)

    def to_markdown(self):
        """
        Generate a summary in markdown format
        """
        json = self.to_report_json()
        # summary = json['summary']

        s = "# Group: {group} - Dataset: {dataset}\n".format(group=json["group"], dataset=json["dataset"])
        s += "\n## SUMMARY\n\n"
        s += "This report generated on {}\n\n".format(datetime.date.today())
        s += "  * Associations: {}\n" . format(json["associations"])
        s += "  * Lines in file (incl headers): {}\n" . format(json["lines"])
        s += "  * Lines skipped: {}\n" . format(json["skipped_lines"])
        # Header from GAF
        s += "## Header From Original Association File\n\n"
        s += "\n".join(["> {}  ".format(head) for head in self.header])
        ## Table of Contents
        s += "\n\n## Contents\n\n"
        for rule, messages in sorted(json["messages"].items(), key=lambda t: t[0]):
            any_suppress_tag_in_rule_metadata = any([tag in self.config.rule_metadata.get(rule, {}).get("tags", []) for tag in self.config.suppress_rule_reporting_tags])
            # For each tag we say to suppress output for, check if it matches any tag in the rule. If any matches
            if self.config.rule_metadata and any_suppress_tag_in_rule_metadata:
                print("Skipping {rule_num} because the tag(s) '{tag}' are suppressed".format(rule_num=rule, tag=", ".join(self.config.suppress_rule_reporting_tags)))
                continue

            s += "[{rule}](#{rule})\n\n".format(rule=rule)

        s += "\n## MESSAGES\n\n"
        for (rule, messages) in sorted(json["messages"].items(), key=lambda t: t[0]):
            any_suppress_tag_in_rule_metadata = any([tag in self.config.rule_metadata.get(rule, {}).get("tags", []) for tag in self.config.suppress_rule_reporting_tags])

            # Skip if the rule metadata has "silent" as a tag
            if self.config.rule_metadata and any_suppress_tag_in_rule_metadata:
                # If there is a rule metadata, and the rule ID is in the config,
                # get the list of tags if present and check for existence of "silent".
                # If contained, continue to the next rule.
                continue


            s += "### {rule}\n\n".format(rule=rule)
            if rule != "other" and self.config.rule_metadata:
                s += "{title}\n\n".format(title=self.config.rule_metadata.get(rule, {}).get("title", ""))
            s += "* total: {amount}\n".format(amount=len(messages))
            if len(messages) > 0:
                s += "#### Messages\n"
            for message in messages:
                obj = " ({})".format(message["obj"]) if message["obj"] else ""
                s += "* {level} - {type}: {message}{obj} -- `{line}`\n".format(level=message["level"], type=message["type"], message=message["message"], line=message["line"], obj=obj)

        return s

@dataclass
class ParseResult:
    parsed_line: str
    associations: List[association.GoAssociation]
    skipped: bool
    report: Optional[Report] = None
    evidence_used: List[str] = None


# TODO avoid using names that are builtin python: file, id

class AssocParser(object):
    """
    Abstract superclass of all association parser classes
    """
    def is_header(self, line):
        return line.startswith("!")

    def parse(self, file, skipheader=False, outfile=None):
        """Parse a line-oriented association file into a list of association dict objects

        Note the returned list is of dict objects. TODO: These will
        later be specified using marshmallow and it should be possible
        to generate objects

        Arguments
        ---------
        file : file or string
            The file is parsed into association objects. Can be a http URL, filename or `file-like-object`, for input assoc file
        outfile : file
            Optional output file in which processed lines are written. This a file or `file-like-object`

        Return
        ------
        list
            Associations generated from the file
        """
        associations = self.association_generator(file, skipheader=skipheader, outfile=outfile)
        a = list(associations)
        return a

    def association_generator(self, file, skipheader=False, outfile=None) -> Dict:
        """
        Returns a generator that yields successive associations from file

        Yields
        ------
        association
        """
        file = self._ensure_file(file)
        for line in file:
            parsed_result = self.parse_line(line)
            self.report.report_parsed_result(parsed_result, outfile, self.config.filtered_evidence_file, self.config.filter_out_evidence)
            for association in parsed_result.associations:
                # yield association if we don't care if it's a header or if it's definitely a real gaf line
                if not skipheader or not isinstance(association, dict):
                    yield association

        logger.info(self.report.short_summary())
        file.close()

    def generate_associations(self, line, outfile=None):
        associations = self.association_generator(line, outfile=outfile)
        for association in associations:
            pass

    def validate_line(self, line):
        if line == "":
            self.report.warning(line, Report.WRONG_NUMBER_OF_COLUMNS, "",
                                msg="GORULE:0000001: empty line", rule=1)
            return ParseResult(line, [], True)

    def map_to_subset(self, file, outfile=None, ontology=None, subset=None, class_map=None, relations=None):
        """
        Map a file to a subset, writing out results

        You can pass either a subset name (e.g. goslim_generic) or a dictionary with ready-made mappings

        Arguments
        ---------
        file: file
            Name or file object for input assoc file
        outfile: file
            Name or file object for output (mapped) assoc file; writes to stdout if not set
        subset: str
            Optional name of subset to map to, e.g. goslim_generic
        class_map: dict
            Mapping between asserted class ids and ids to map to. Many to many
        ontology: `Ontology`
            Ontology to extract subset from

        """
        if subset is not None:
            logger.info("Creating mapping for subset: {}".format(subset))
            class_map = ontology.create_slim_mapping(subset=subset, relations=relations)

        if class_map is None:
            raise ValueError("Neither class_map not subset is set")
        col = self.ANNOTATION_CLASS_COLUMN
        file = self._ensure_file(file)
        tuples = []
        for line in file:
            if line.startswith("!"):
                continue
            vals = line.split("\t")
            logger.debug("LINE: {} VALS: {}".format(line, vals))
            if len(vals) < col:
                raise ValueError("Line: {} has too few cols, expect class id in col {}".format(line, col))
            cid = vals[col]
            if cid not in class_map or len(class_map[cid]) == 0:
                self.report.error(line, Report.UNMAPPED_ID, cid)
                continue
            else:
                for mcid in class_map[cid]:
                    vals[col] = mcid
                    line = "\t".join(vals)
                    if outfile is not None:
                        outfile.write(line)
                    else:
                        print(line)


    def skim(self, file):
        """
        Lightweight parse of a file into tuples.

        Note this discards metadata such as evidence.

        Return a list of tuples (subject_id, subject_label, object_id)
        """
        raise NotImplementedError("AssocParser.skim not implemented")

    def normalize_columns(self, number_of_columns, columns):
        columns += [""] * (number_of_columns - len(columns))
        return columns

    # def parse_line(self, line) -> ParseResult:
    #     if self.is_header(line):
    #         # Then do a thing with the header?
    #         return ParseResult(line, [HeaderLine(line)], False)
    #
    #     tsv_line = line.split("\t")
    #     self.parse_to_association(tsv_line)


    def _skipping_line(self, associations):
        return associations is None or associations == []

    def _is_exclude_relation(self, relation):
        if self.config.include_relations is not None and len(self.config.include_relations) > 0:
            if relation not in self.config.include_relations:
                return True
        if self.config.exclude_relations is not None and len(self.config.exclude_relations) > 0:
            if relation in self.config.exclude_relations:
                return True
        return False

    def compute_aspect(self, term):
        if self.config.ontology == None:
            return None

        BP = "GO:0008150"
        CC = "GO:0005575"
        MF = "GO:0003674"

        ancestors = self.config.ontology.ancestors(term)
        if BP in ancestors:
            return "P"
        if CC in ancestors:
            return "C"
        if MF in ancestors:
            return "F"

        return None

    ## we generate both qualifier and relation field
    ## Returns: (negated, relation, other_qualifiers)
    def _parse_qualifier(self, qualifier, aspect):
        return _parse_qualifier(qualifier, aspect)

    # split an ID/CURIE into prefix and local parts
    # (not currently used)
    def _parse_id(self, id):
        toks = id.split(":")
        if len(toks) == 2:
            return (toks[0],toks[1])
        else:
            return (toks[0],toks[1:].join(":"))

    # split an ID/CURIE into prefix and local parts
    def _get_id_prefix(self, id):
        toks = id.split(":")
        return toks[0]

    def _validate_taxon(self, taxon, line: SplitLine):
        if self.config.valid_taxa is None:
            return True
        else:
            if taxon in self.config.valid_taxa:
                return True
            else:
                self.report.error(line.line, Report.INVALID_TAXON, taxon, taxon=taxon)
                return False

    # check the term id is in the ontology, and is not obsolete
    def _validate_ontology_class_id(self, id, line: SplitLine, subclassof=None):
        ont = self.config.ontology
        if ont is None:
            return id

        if not ont.has_node(id):
            self.report.warning(line.line, Report.UNKNOWN_ID, id, "Class ID {} is not present in the ontology".format(id),
                taxon=line.taxon, rule=27)
            return id

        if ont.is_obsolete(id):
            # the default behavior should always be to repair, unless the caller explicitly states
            # that this should not be done by setting repair_obsoletes to False
            if self.config.repair_obsoletes is None or self.config.repair_obsoletes:
                rb = ont.replaced_by(id, strict=False)
                if len(rb) == 1:
                    # We can repair
                    self.report.warning(line.line, Report.OBSOLETE_CLASS, id, msg="Violates GORULE:0000020, but was repaired",
                        taxon=line.taxon, rule=20)
                    id = rb[0]
                else:
                    self.report.warning(line.line, Report.OBSOLETE_CLASS_NO_REPLACEMENT, id, msg="Violates GORULE:0000020",
                        taxon=line.taxon, rule=20)
                    id = None
            else:
                self.report.warning(line.line, Report.OBSOLETE_CLASS, id, msg="Violates GORULE:0000020",
                    taxon=line.taxon, rule=20)
                id = None

        return id

    def _validate_symbol(self, symbol, line: SplitLine):
        if symbol is None or symbol == "":
            self.report.warning(line.line, Report.INVALID_SYMBOL, symbol, "GORULE:0000027: symbol is empty",
                taxon=line.taxon, rule=27)

    non_id_regex = re.compile(r"[^\.:_\-0-9a-zA-Z]")
    doi_regex = re.compile(r"^doi:", flags=re.IGNORECASE)

    def _validate_id(self, id, line: SplitLine, allowed_ids=None, context=None):

        # we assume that cardinality>1 fields have been split prior to this

        if id == "":
            self.report.error(line.line, Report.INVALID_ID, id, "GORULE:0000027: identifier is empty", taxon=line.taxon, rule=27)
            return False
        if ':' not in id:
            self.report.error(line.line, Report.INVALID_ID, id, "GORULE:0000027: must be CURIE/prefixed ID", rule=27)
            return False

        # we won't check IDs with doi prefix, everything else we want to check
        if not AssocParser.doi_regex.match(id) and AssocParser.non_id_regex.search(id):
            self.report.error(line.line, Report.INVALID_ID, id, "GORULE:0000027: contains non letter, non number character, or spaces", rule=27)
            return False

        (id_prefix, right) = id.split(":", maxsplit=1)
        if right.startswith("MGI:"):
            ## See ticket https://github.com/geneontology/go-site/issues/91
            ## For purposes of determining allowed IDs in DB XREF, MGI IDs shall look like `MGI:12345`
            right = right[4:]

        if id_prefix == "" or right == "":
            self.report.error(line.line, Report.INVALID_ID, id, "GORULE:0000027: Empty ID", rule=27)
            return False

        if allowed_ids is not None and id_prefix not in allowed_ids:
            # For now we will just issue a warning here, and we won't filter out the annotation here
            self.report.warning(line.line, Report.INVALID_ID_DBXREF, id_prefix, "allowed: {}".format(allowed_ids), rule=27)

        # ensure that the ID space of the annotation class (e.g. GO)
        # conforms to what is expected
        if context == ANNOTATION and self.config.class_idspaces is not None:
            if id_prefix not in self.config.class_idspaces:
                self.report.error(line.line, Report.INVALID_IDSPACE, id_prefix, "allowed: {}".format(self.config.class_idspaces), rule=27)
                return False

        return True

    def validate_pipe_separated_ids(self, column, line: SplitLine, empty_allowed=False, extra_delims="") -> Optional[List[str]]:
        if column == "" and empty_allowed:
            return []

        split_ids = re.split("[|{}]".format(extra_delims), column)
        valids = []
        for i in split_ids:
            if self._validate_id(i, line):
                valids.append(i)
            else:
                return None

        return sorted(valids)

    def validate_curie_ids(self, ids: List[association.Curie], line: SplitLine) -> Optional[List[association.Curie]]:
        valids = []
        for i in ids:
            if self._validate_id(str(i), line):
                valids.append(i)
            else:
                return None

        return valids

        # We are only reporting, so just pass it through

    def _normalize_id(self, id):
        toks = id.split(":")
        if len(toks) > 1:
            return self._pair_to_id(toks[0], ":".join(toks[1:]))
        else:
            return id

    def _pair_to_id(self, db, localid):
        if self.config.remove_double_prefixes:
            ## Switch MGI:MGI:n to MGI:n
            if localid.startswith(db+":"):
                localid = localid.replace(db+":", "")
        return db + ":" + localid

    def _taxon_id(self, id, line: SplitLine):
        id = id.replace('taxon', 'NCBITaxon')
        valid = self._validate_id(id, line, context=TAXON)
        if valid:
            return id
        else:
            return None

    def _ensure_file(self, file):
        logger.info("Ensure file: {}".format(file))
        if isinstance(file,str):
            # TODO Let's fix this if/elseif chain.
            if file.startswith("ftp"):
                f = tempfile.NamedTemporaryFile()
                fn = f.name
                cmd = ['wget',file,'-O',fn]
                subprocess.run(cmd, check=True)
                return open(fn,"r")
            elif file.startswith("http"):
                url = file
                with closing(requests.get(url, stream=False, headers={'User-Agent': get_user_agent(modules=[requests], caller_name=__name__)})) as resp:
                    logger.info("URL: {} STATUS: {} ".format(url, resp.status_code))
                    ok = resp.status_code == 200
                    if ok:
                        logger.debug("HEADER: {}".format(resp.headers))
                        if file.endswith(".gz"):
                            return io.StringIO(str(gzip.decompress(resp.content),'utf-8'))
                        else:
                            out = io.StringIO(resp.content)
                            return out
                    else:
                        return None
            else:
                logger.info("Testing suffix of {}".format(file))
                if file.endswith(".gz"):
                    return gzip.open(file, "rt")
                else:
                    return open(file, "r")
        else:
            return file

    def _parse_full_extension_expression(self, xp, line: SplitLine = None):
        if xp == "":
            return None

        object_or_exprs = []
        xp_ors = sorted(xp.split("|"))
        for xp_or in xp_ors:

            # gather conjunctive expressions in extensions field
            xp_ands = sorted(xp_or.split(","))
            and_exprs = []
            for xp_and in xp_ands:
                if xp_and != "":
                    expr = self._parse_relationship_expression(xp_and, line=line)
                    if expr is not None:
                        and_exprs.append(expr)
            if len(and_exprs) > 0:
                object_or_exprs.append({'intersection_of': and_exprs})
        return object_or_exprs


    relation_tuple = re.compile(r'(.+)\((.+)\)')
    def _parse_relationship_expression(self, x, line: SplitLine = None):
        ## Parses an atomic relational expression
        ## E.g. exists_during(GO:0000753)
        ## Atomic class expressions only
        tuples = AssocParser.relation_tuple.findall(x)
        if len(tuples) != 1:
            self.report.error(line.line, Report.EXTENSION_SYNTAX_ERROR, x, msg="does not follow REL(ID) syntax")
            return None
        (p, v) = tuples[0]

        if self._validate_id(v, line, context=EXTENSION):
            return {
                'property': p,
                'filler': v
            }
        else:
            self.report.error(line.line, Report.EXTENSION_SYNTAX_ERROR, x, msg="GORULE:0000027: ID not valid", rule=27)
            return None



def parse_date(date: str, report: Report, line: List) -> Optional[association.Date]:
    if date == "":
        report.error(line, Report.INVALID_DATE, "\'\'", "GORULE:0000001: empty", rule=1)
        return None

    d = None
    if len(date) == 8 and date.isdigit():
        # For GAF date, should be YYYYMMDD all as digits and
        # a well formed date string here will be exactly 8 characters long
        d = association.Date(year=date[0:4], month=date[4:6], day=date[6:8], time="")
    else:
        report.warning(line, report.INVALID_DATE, date, "GORULE:0000001: Date field must be YYYYMMDD, got: {}".format(date), rule=1)
        parsed = None
        try:
            parsed = dateutil.parser.parse(date)
        except:
            report.error(line, Report.INVALID_DATE, date, "GORULE:0000001: Could not parse date '{}' at all".format(date), rule=1)
            return None

        d = association.Date(year="{:02d}".format(parsed.year), month="{:02d}".format(parsed.month), day="{:02d}".format(parsed.day), time="")

    return d

def parse_iso_date(date: str, report: Report, line: List) -> Optional[association.Date]:

    def parse_with_dateutil(date: str, repot: Report, line: List) -> Optional[association.Date]:
        parsed = None
        try:
            parsed = dateutil.parser.parse(date)
        except:
            report.error(line, Report.INVALID_DATE, date, "GORULE:0000001: Could not parse date '{}' at all".format(date), rule=1)
            return None

        return association.Date(
            year="{:02d}".format(parsed.year),
            month="{:02d}".format(parsed.month),
            day="{:02d}".format(parsed.day),
            time=parsed.time().isoformat())


    if date == "":
        report.error(line, Report.INVALID_DATE, "\'\'", "GORULE:0000001: empty", rule=1)
        return None

    d = None
    if len(date) >= 10:
        # For ISO style date, should be YYYY-MM-DD all as digits and
        # a well formed date string here will be at least 10 characters long.
        # Optionally, there could be an appended THH:MM
        year = date[0:4]
        month = date[5:7]
        day = date[8:10]
        time = date[11:]
        if year.isdigit() and month.isdigit() and day.isdigit():
            d = association.Date(year=year, month=month, day=day, time=time)
        else:
            # If any of year, month, day are not made of digits, then try to parse with dateutil
            report.warning(line, report.INVALID_DATE, date, "GORULE:0000001: Date field must be YYYY-MM-DD, got: {}".format(date), rule=1)
            d = parse_with_dateutil(date, report, line)

    else:
        # If we got the wrong number characters in the date, then fallback to dateutil
        report.warning(line, report.INVALID_DATE, date, "GORULE:0000001: Date field must be YYYY-MM-DD, got: {}".format(date), rule=1)
        d = parse_with_dateutil(date, report, line)

    return d


## we generate both qualifier and relation field
## Returns: (negated, relation, other_qualifiers)
def _parse_qualifier(qualifier, aspect):
    relation = None
    qualifiers = qualifier.split("|")
    if qualifier == '':
        qualifiers = []
    negated = 'NOT' in qualifiers
    other_qualifiers = [q for q in qualifiers if q != 'NOT']
    ## In GAFs, relation is overloaded into qualifier.
    ## If no explicit non-NOT qualifier is specified, use
    ## a default based on GPI spec
    if len(other_qualifiers) > 0:
        relation = other_qualifiers[0]
    else:
        if aspect == 'C':
            relation = 'part_of'
        elif aspect == 'P':
            relation = 'involved_in'
        elif aspect == 'F':
            relation = 'enables'
        else:
            relation = None
    return (negated, relation, other_qualifiers)

# TODO consider making an Association its own class too to give it a little more
# TODO Semantic value?

# TODO consider making an ID class?



from ontobio import OntologyFactory
from ontobio.tsv_expander import expand_tsv
import csv


INPUT = "tests/resources/data_table.tsv"
OUTPUT = "tests/resources/data_table-expanded.tsv"

def test_expand():
    factory = OntologyFactory()
    ontobj = factory.create("tests/resources/goslim_pombe.json")
    expand_tsv(INPUT, ontology=ontobj, outfile=open(OUTPUT,"w"), cols=["term"])
    reader = csv.DictReader(open(OUTPUT, "r"), delimiter='\t')
    n=0
    for row in reader:
        if row['term'] == 'GO:0002181':
            assert row['term_label'] == 'cytoplasmic translation'
            n += 1
        if row['term'] == 'FAKE:123':
            assert row['term_label'] == ''
            n += 1
    assert n == 2
